// Graph.cpp : implementation file
//

#include "stdafx.h"
#include "Graph.h"
#include "Colors.h"

//#define DEBUG_SCALING
#ifdef DEBUG_SCALING
#define GRAPHTRACE(x) TRACE x
#else
#define GRAPHTRACE(x)
#endif

//#define DEBUG_MONITOR
#ifdef DEBUG_MONITOR
#define MONITORTRACE(x) TRACE x
#else
#define MONITORTRACE(x)
#endif

// CGraph

IMPLEMENT_DYNAMIC(CGraph, CStatic)
CGraph::CGraph()
   {
    CanPlot = FALSE;
    BB = CPoint(0,0);
    info = NULL;
   }

CGraph::~CGraph()
{
}


/****************************************************************************
*                                 Message Map
****************************************************************************/

BEGIN_MESSAGE_MAP(CGraph, CStatic)
    ON_WM_PAINT()
    ON_WM_ERASEBKGND()
    ON_WM_DESTROY()
    ON_WM_MOUSEMOVE()
    ON_MESSAGE(WM_MOUSELEAVE, OnMouseLeave)
    ON_WM_SIZE()
END_MESSAGE_MAP()

/****************************************************************************
*                               CGraph::Create
* Inputs:
*       const CRect & r: Rectangle to create it in
*       CWnd * parent: Parent graph
* Result: BOOL
*       TRUE if successful
*       FALSE if error
* Effect: 
*       Creates the graph window
****************************************************************************/

BOOL CGraph::Create(const CRect & r, CWnd * parent)
    {
     if(!CStatic::Create(NULL, WS_CHILD | WS_VISIBLE | SS_NOTIFY, r, parent))
        return FALSE;
     return TRUE;
    } // CGraph::Create

/****************************************************************************
*                           CGraph::SetMappingMode
* Inputs:
*       CDC & dc: Display context
* Result: BOOL
*       TRUE if it could be set
*       FALSE if there is not enough information to set it
* Effect: 
*       Sets the mapping mode for the DC based on the winodw size and
*       the bounding box
****************************************************************************/

BOOL CGraph::SetMappingMode(CDC & dc)
    {
     CRect r;
     GetClientRect(&r);

     GRAPHTRACE( ("BB=(%d, %d)\n", BB.x, BB.y) );
     if(BB.x == 0 && BB.y == 0)
        return FALSE;

     dc.SetMapMode(MM_ANISOTROPIC);

     CSize vp;
#define INITIAL_SCALING_FACTOR 10.0
     double factor = INITIAL_SCALING_FACTOR;

     GRAPHTRACE( ("r=(%d, %d)\n", r.Width(), r.Height()) );

     //****************************************************************
     // In attempting to normalize the scaling factors, it is possible
     // that roundoff error will cause the values to go to 0 or to some
     // rather small value like 1 or 2.  This either causes the
     // information to disappear immediately or to scale very strangely
     // during resizing operations.  So the goal here is to get a
     // scaling factor of at least MIN_SCALING_FACTOR
     // We encourage this by setting the 'factor' to INITIAL_SCALING_FACTOR
     // to start with
     //****************************************************************
#define MIN_SCALING_FACTOR 10
     while(TRUE)
        { /* compute magnifier */ 
         vp = CSize((int) ((double)r.Width() * (factor * (double)r.Width())/(double)BB.x),
                    (int) ((double)r.Height() * (factor * (double)r.Height())/(double)BB.y));

         GRAPHTRACE( ("vp=(%d, %d) factor = %g\n", vp.cx, vp.cy, factor));
         if(vp.cx < MIN_SCALING_FACTOR || vp.cy < MIN_SCALING_FACTOR)
            { /* remultiply */
             factor *= 2.0;
             continue;
            } /* remultiply */
         break;
        } /* compute magnifier */

     dc.SetWindowOrg(0, BB.y);
     GRAPHTRACE( ("SetWindowOrg(%d, %d)\n", 0, BB.y));

     CSize we((int) (factor * (double)r.Width()), (int)(factor * (double)-r.Height()));
     GRAPHTRACE(("SetWindowExt(%d, %d)\n", we.cx, we.cy));
     dc.SetWindowExt(we);

     GRAPHTRACE(("SetViewportExt(%d, %d)\n", vp.cx, vp.cy));
     dc.SetViewportExt(vp);
     
     dc.SetViewportOrg(0,0);
     return TRUE;
    } // CGraph::SetMappingMode

/****************************************************************************
*                               CGraph::OnPaint
* Result: void
*       
* Effect: 
*       Draws the graph data
****************************************************************************/

void CGraph::OnPaint()
    {
     CPaintDC dc(this); // device context for painting

     CRect r;
     GetClientRect(&r);
     dc.DrawEdge(&r, EDGE_SUNKEN, BF_ADJUST | BF_RECT);
     if(!CanPlot)
        { /* not ready */
         dc.FillSolidRect(&r, ::GetSysColor(COLOR_3DFACE));
         return;
        } /* not ready */
     else
        dc.FillSolidRect(&r, RGB(128, 128, 0));

     //****************************************************************
     // Because the clipping region of a CStatic is actually based on
     // the CS_PARENTDC style, and goes beyond the edges, it is possible
     // to "paint outside the lines".  By selecting a clipping region
     // limited to the (adjusted) client rectangle, we limit how much
     // it can paint
     //****************************************************************
     CRgn rgn;
     rgn.CreateRectRgn(r.left, r.top, r.right, r.bottom);

     int save = dc.SaveDC();

     dc.SelectClipRgn(&rgn);
     
     if(SetMappingMode(dc))
        { /* has bounding box */
         for(int i = 0; i < MAX_THREADS; i++)
            lines[i].Plot(dc);
        } /* has bounding box */

     dc.RestoreDC(save);
     // The RestoreDC makes sure the CRgn is not selected into a DC
     // at the time when the CRgn::~CRgn is invoked.
     
     // Do not call CStatic::OnPaint() for painting messages
    }

/****************************************************************************
*                           CGraph::LineInfo::Plot
* Inputs:
*       CDC & dc: A prepped DC
* Result: void
*       
* Effect: 
*       Plots all the lines
****************************************************************************/
#define PEN_WIDTH_PHYSICAL 2

void CGraph::LineInfo::Plot(CDC & dc)
    {
     if(points.GetSize() == 0)
        return;
     // We want an n-point-wide pen, physically, but make sure of this
     // by creating a pen in logical coordinates that is at least
     // n physical units
     CSize penwidth(PEN_WIDTH_PHYSICAL, PEN_WIDTH_PHYSICAL);

//#define DEBUG_TOUCH
#ifdef DEBUG_TOUCH  // debugging mode
     ConstructRegion(GetFuzz(dc));
     if((HRGN)boundingRgn != NULL)
        { /* has bouuding */
         // show the bounding region
         CBrush br(RGB(128, 128, 128));
         dc.FillRgn(&boundingRgn, &br);
         CBrush blk(RGB(0,0,0));
         dc.FrameRgn(&boundingRgn, &blk, 1,1);
        } /* has bouuding */
#endif

//     dc.DPtoLP(&penwidth);
//     TRACE("PenWidth=%d\n", penwidth.cx);
     
     CPen pen(PS_SOLID, penwidth.cx, PriorityColors::GetBkColor(priority));

     int save = dc.SaveDC();
         dc.SelectObject(&pen);
         dc.Polyline(points.GetData(), (int)points.GetSize());
     dc.RestoreDC(save);
    } // CGraph::LineInfo::Plot

/****************************************************************************
*                                CGraph::Clear
* Result: void
*       
* Effect: 
*       Deletes all data points and clears the graph
****************************************************************************/

void CGraph::Clear()
    {
     // Reset the bounding box
     BB.x = 0;
     BB.y = 0;
     
     // clear all data points
     for(int i = 0; i < MAX_THREADS; i++)
         lines[i].points.RemoveAll();
         
     CanPlot = FALSE;

     // Now force a redraw
     Invalidate();
    } // CGraph::Clear

/****************************************************************************
*                            CGraph::OnEraseBkgnd
* Inputs:
*       CDC * pDC:
* Result: BOOL
*       TRUE, always
* Effect: 
*       Actually, does nothing, forcing all erasure into the OnPaint handler
****************************************************************************/

BOOL CGraph::OnEraseBkgnd(CDC* pDC)
    {
     return TRUE; 
    }

/****************************************************************************
*                              CGraph::AddPoint
* Inputs:
*       INT_PTR index: What to add the point to
*       int priority: Thread priority code
*       CPoint pt: Poit to plot
* Result: void
*       
* Effect: 
*       Adds the point to the plot.  Adjusts x,y limits
****************************************************************************/

void CGraph::AddPoint(INT_PTR index, int priority, CPoint pt)
    {
     // Note that we currently only allow one color for the line
     lines[index].priority = priority;
     lines[index].points.Add(pt); 
     BB.x = max(BB.x, pt.x);
     BB.y = max(BB.y, pt.y);
    } // CGraph::AddPoint

/****************************************************************************
*                                CGraph::Plot
* Result: void
*       
* Effect: 
*       Plots all the points based on the currently-computed BB for scaling
****************************************************************************/

void CGraph::Plot()
    {
     CanPlot = TRUE;
     Invalidate();
    } // CGraph::Plot

/****************************************************************************
*                              CGraph::OnDestroy
* Result: void
*       
* Effect: 
*       Clears any data left in the graph
****************************************************************************/

void CGraph::OnDestroy()
    {
     CStatic::OnDestroy();

     Clear();
    }

/****************************************************************************
*                               CGraph::GetFuzz
* Inputs:
*       CDC & dc: Display context with mapping mode set
* Result: int
*       The logical fuzz factor
****************************************************************************/
#define FUZZFACTOR 5

/* static */ int CGraph::GetFuzz(CDC & dc)
    {
     CSize fuzz(FUZZFACTOR, FUZZFACTOR);
     dc.DPtoLP(&fuzz);
     return fuzz.cx;
    } // CGraph::GetFuzz

/****************************************************************************
*                             CGraph::OnMouseMove
* Inputs:
*       UINT nFlags:
*       CPoint point:
* Result: void
*       
* Effect: 
*       Pops up the info display
****************************************************************************/

void CGraph::OnMouseMove(UINT nFlags, CPoint point)
    {
     CClientDC dc(this);
     SetMappingMode(dc);
     CPoint mp = point; // mapped point
     dc.DPtoLP(&mp);

     int fuzz = GetFuzz(dc);

     for(int i = 0; i < MAX_THREADS; i++)
        { /* check each */
         if(lines[i].InBoundingRgn(mp, fuzz))
            { /* got it */
             CString s;
             s = lines[i].GetText();
             SetInfoWindow(point, s, TRUE);
             return;
            } /* got it */
        } /* check each */

     SetInfoWindow(point, graphInfo.GetText(), TRUE);

     CStatic::OnMouseMove(nFlags, point);
    }

/****************************************************************************
*                          CGraph::PositionOnMonitor
* Result: BOOL
*       TRUE if successful
*       FALSE if failed
* Effect: 
*       Tries to position the window correctly on multiple monitors
****************************************************************************/

BOOL CGraph::PositionOnMonitor(CPoint target)
    {
     HMONITOR monitor = ::MonitorFromPoint(target, MONITOR_DEFAULTTONEAREST);
     if(monitor == NULL)
        return FALSE;

     // The goal here is to keep the window from falling off the right end of
     // a monitor by repositioning the text to the left, as long as this
     // doesn't force the window off the left edge of the current monitor
     // All the various boundary conditions would be very complex, so this
     // is a simple test
     // Adjustment cases:
     //
     // Left origin at target fits in window
     // [A]
     //      +--------------------------+      +---------------------------+
     //      |   *                      |      |   *                       |
     //      |   +----------------+     |      |   +----------------+      | 
     //      |   |                |     |  ==> |   |                |      |
     //      |   +----------------+     |      |   +----------------+      | 
     //      |                          |      |                           |
     //      +--------------------------+      +---------------------------+
     //
     // Window is wider than monitor.
     // [B]                                              [B.1]
     //      +--------------------------+                +---------------------------+
     //      |   *                      |                |   *                       |
     //      |   +--------------------------------+      |+--------------------------------+      
     //      |   |                                |  ==> ||                                |
     //      |   +--------------------------------+      |+--------------------------------+
     //      |                          |                |                           |
     //      +--------------------------+                +---------------------------+
     //
     // The correct behavior in this case depends on whether the secondary monitor is to the left, right, or vertical
     // relative to the monitor that contains the window.
     //         B.1 If the monitors are vertical, the above solution is the only one that can be done
     //         B.2 If the secondary monitor is to the right, B.1. also works fine
     //         B.3 If the secondary monitor is to the left, the following must be implemented
     //
     //           [B.3]
     //  +---------------------------+
     //           |   *                       |
     //     +--------------------------------+|      
     // ==> |                                ||
     //     +--------------------------------+|      
     //           |                           |
     //  +---------------------------+
     //
     // Overhangs on right, but when adjusted to be left of target, fits
     // [C]
     //                                        [C.1]
     //      +--------------------------+      +---------------------------+
     //      |               *          |      |                 *         |
     //      |               +---------------+ | +---------------+         |
     //      |               |               | | |               |         |
     //      |               +---------------+ | +---------------+         |
     //      |                          |      |                           |
     //      +--------------------------+      +---------------------------+
     // alternatively, this would work, and is less visually disturbing
     // because the window just "locks" onto the edge of the window instead
     // of flip-flopping back and forth
     //                                        [C.2]
     //                                        +---------------------------+
     //                                        |                 *         |
     //                                        |          +---------------+|
     //                                        |          |               ||
     //                                        |          +---------------+|
     //                                        |                           |
     //                                        +---------------------------+
     //                                        
     //
     // Overhangs on right, but when adjusted to be left of target, overhangs on left
     // Note that this can be subsumed into case C.2
     // [D]                                                  [D.1]
     //      +-----------------------------+                 +-----------------------------+
     //      |                *            |                 |                 *           |
     //      |                +--------------------+      +--------------------+           |
     //      |                |                    | ==>  |                    |           |
     //      |                +--------------------+      +--------------------+           |
     //      |                             |                 |                             |
     //      +-----------------------------+                 +-----------------------------+
     //
     //      [D.2]
     //      +-----------------------------+
     //      |            *                |
     //      |+--------------------+       |
     //  ==> ||                    |       |
     //      |+--------------------+       |
     //      |                             |
     //      +-----------------------------+

     MONITORINFO moninfo = { sizeof(MONITORINFO) };
     if(!::GetMonitorInfo(monitor, &moninfo))
        return FALSE;

     // Make it a CRect so we can easily get Width(), Height()
     CRect rcMonitor = moninfo.rcMonitor;

     CRect w;
     info->GetWindowRect(&w);
     
     if(target.x + w.Width() <= rcMonitor.right)
        { /* [A] */
         MONITORTRACE( ("PositionOnMonitor(%d, %d) width %d => [A] (%d, %d)\n", 
                         target.x, target.y, w.Width(), target.x, target.y) );
         info->SetWindowPos(NULL, target.x, target.y,
                            0, 0,
                            SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
         return TRUE;
        } /* [A] */

     if(w.Width() > rcMonitor.Width())
        { /* B */
         // is there a monitor to the right?
         CPoint inRight(rcMonitor.left + w.Width(), target.y);
         HMONITOR toright = ::MonitorFromPoint(inRight, MONITOR_DEFAULTTONULL);
         if(toright != NULL)
            { /* B.2 */
             MONITORTRACE( ("PositionOnMonitor(%d, %d) width %d => [B.2] (%d, %d)\n",
                             target.x, target.y, w.Width(), rcMonitor.left, target.y) );
             info->SetWindowPos(NULL, rcMonitor.left, target.y,
                               0, 0,
                               SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
             return TRUE;
            } /* B.2 */
         CPoint inLeft(rcMonitor.right - w.Width(), target.y);
         HMONITOR toleft = ::MonitorFromPoint(inLeft, MONITOR_DEFAULTTONULL);
         if(toleft != NULL)
            { /* B.3 */
             MONITORTRACE( ("PositionOnMonitor(%d, %d) width %d => [B.3] (%d, %d)\n",
                             target.x, target.y, w.Width(), rcMonitor.right - w.Width(), target.y) );
             info->SetWindowPos(NULL, rcMonitor.right - w.Width(), target.y,
                                0, 0,
                                SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
             return TRUE;
            } /* B.3 */
         // Must be B.1.  We lose,but favor maximum information from the left
         MONITORTRACE( ("PositionOnMonitor(%d, %d) width %d => [B.1] (%d, %d)\n",
                         target.x, target.y, w.Width(), rcMonitor.left, target.y) );
         info->SetWindowPos(NULL, rcMonitor.left, target.y,
                            0, 0,
                            SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
         return TRUE;
        } /* B */

     if(target.x + w.Width() > rcMonitor.right)
        { /* C,D */
         // Favor putting on one screen over letting it hang across
         // two screens
         // The two options are as show, C.1 or C.2.  C2 is less
         // visually disruptive (I tried C1 and it was ugly)
         // This also subsumes case D as a special case
         MONITORTRACE( ("PositionOnMonitor(%d, %d) width %d => [C,D] (%d, %d)\n",
                         target.x, target.y, w.Width(), rcMonitor.right - w.Width(), target.y) );
         info->SetWindowPos(NULL, rcMonitor.right - w.Width(), target.y,
                            0, 0, 
                            SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
         return TRUE;
        } /* C,D */
     return FALSE;
    } // CGraph::PositionOnMonitor

/****************************************************************************
*                            CGraph::SetInfoWindow
* Inputs:
*       CPoint point: Coordinates of mouse, in Client coordinates
*       const CString & text:
*       BOOL visible:
* Result: void
*       
* Effect: 
*       Pops up the info window, or hides it
****************************************************************************/

void CGraph::SetInfoWindow(CPoint point, const CString & text, BOOL visible)
    {
     TRACE("SetInfoWindow( (%d, %d), \"%s\", %s)\n", point.x, point.y, text, visible ? _T("true") : _T("false"));
     CPoint target = point;
     ClientToScreen(&target);
     target.y += ::GetSystemMetrics(SM_CYCURSOR);

     if(info != NULL)
        { /* move window */
         info->SetWindowText(text);
        } /* move window */
     else
        { /* create window */
         info = new CInfoDisplay;
         if(!info->Create(target.x, target.y, text, this))
            { /* failed */
             return;
            } /* failed */
         TRACKMOUSEEVENT tme = {sizeof(TRACKMOUSEEVENT), TME_LEAVE, m_hWnd, 0};
         VERIFY(TrackMouseEvent(&tme));
        } /* create window */

     if(!PositionOnMonitor(target))
        { /* failed, use default */
         MONITORTRACE( ("PositionOnMonitor(%d, %d) failed => default to (%d, %d)\n",
                         target.x, target.y, target.x, target.y) );
         
         info->SetWindowPos(NULL, target.x, target.y,
                           0, 0,
                           SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
        } /* failed, use default */
     info->ShowWindow(visible ? SW_SHOWNA : SW_HIDE);
    } // CGraph::SetInfoWindow

/****************************************************************************
*                            CGraph::OnMouseLeave
* Inputs:
*       WPARAM: ignored
*       LPARAM: ignored
* Result: LRESULT
*       Logically void, 0, always
* Effect: 
*       Handles the mouse leaving the window
****************************************************************************/

LRESULT CGraph::OnMouseLeave(WPARAM, LPARAM)
    {
     if(info != NULL)
         info->DestroyWindow();
     info = NULL;
     return 0;
    } // CGraph::OnMouseLeave

/****************************************************************************
*                               CGraph::OnSize
* Inputs:
*       UINT nType: Type of resize
*       int cx: New client width
*       int cy: New client height
* Result: void
*       
* Effect: 
*       Deletes any bounding boxes that might have existed
****************************************************************************/

void CGraph::OnSize(UINT nType, int cx, int cy)
    {
     CStatic::OnSize(nType, cx, cy);

     for(int i = 0; i < MAX_THREADS; i++)
        { /* kill all */
         lines[i].ResetBoundingRgn();
        } /* kill all */
    }

/****************************************************************************
*                       CGraph::LineInfo::InBoundingRgn
* Inputs:
*       CPoint pt: Point in logical coordinates
*       int fuzz: Fuzz factor expressed in logical coordinates
* Result: BOOL
*       TRUE if the mouse is in the bounding region
*       FALSE if it isn't
* Effect: 
*       If there is a need, the bounding region is created
* Notes:
*       The coordinate system is MM_ANISOTROPIC with 0,0 at the
*       bottom left and y increasing upwards
*       ^
*       |
*       |
*       +------>
****************************************************************************/

BOOL CGraph::LineInfo::InBoundingRgn(CPoint pt, int fuzz)
    {
     for(int i = 0; i < MAX_THREADS; i++)
        { /* filter points */
         if(points.GetSize() == 0)
            continue;  // no points in this line

         //----------------------------------------------------------------
         //        /
         //  *A   /      *B
         //      /
         // ____/
         //----------------------------------------------------------------
         if(pt.x < points[0].x)
            continue; // in this graph, all lines start at 0,0, so this shouldn't occur
         if(pt.y < points[0].y)
            continue; // in this graph, all lines start at 0,0, so this shouldn't occur
         if(pt.x > points[points.GetSize() - 1].x)
            continue; // we are in region B
         if(pt.y > points[points.GetSize() - 1].y)
            continue; // in this graph, all lines run to maxy, so this shouldn't occur

         // Make sure the region actually exists
         ConstructRegion(fuzz);

         if((HRGN)boundingRgn != NULL)
            { /* test containment */
             BOOL result = boundingRgn.PtInRegion(pt);
             TRACE("(%d, %d) in region\n", pt.x, pt.y);
             return result;
            } /* test containment */
         return FALSE;
        } /* filter points */
    return FALSE;
    } // CGraph::LineInfo::InBoundingRgn

/****************************************************************************
*                       CGraph::LineInfo::ConstructRegion
* Inputs:
*       int fuzz: Fuzz factor expressed in logical coordinates
* Result: BOOL
*       TRUE if region can be constructed
*       FALSE if error
* Effect: 
*       Constructs a CRgn object
* Notes:
*       The current CRgn is presumed to be NULL
*
*              2*-2*-*3
*              /  / /
*             /  / /
*           1* 1* *4
*           /  / /
*          /  / /
*        0* 0* *5
*        ^  ^  ^
*        |  |  |
*        |  |  outer bound (poly indices)
*        |  line (lines indices)
*        inner bound (poly indices)
****************************************************************************/

BOOL CGraph::LineInfo::ConstructRegion(int fuzz)
    {
     if( (HRGN)boundingRgn != NULL)
         return TRUE;  // already exists
     TRACE("Constructing Polygon Region\n");
     CArray<CPoint, CPoint>poly;
     poly.SetSize(points.GetSize() * 2);
     for(int i = 0; i < points.GetSize(); i++)
        { /* compute points */
         INT_PTR k = (2 * points.GetSize()) - 1 - i;
         poly[i].x = points[i].x - fuzz;
         poly[i].y = points[i].y;
         poly[k].x = points[i].x + fuzz;
         poly[k].y = points[i].y;
        } /* compute points */
     
     if(!boundingRgn.CreatePolygonRgn(poly.GetData(), (int)poly.GetSize(), ALTERNATE))
        { /* creation failed */
         ASSERT(FALSE);
         return FALSE;
        } /* creation failed */
     return TRUE;
    } // CGraph::LineInfo::ConstructRegion

/****************************************************************************
*                     CGraph::LineInfo::ResetBoundingRgn
* Result: void
*       
* Effect: 
*       Clears the bounding region
****************************************************************************/

void CGraph::LineInfo::ResetBoundingRgn()
    {
     boundingRgn.DeleteObject();
    } // CGraph::LineInfo::ResetBoundingRgn

/****************************************************************************
*                            CGraph::SetThreadInfo
* Inputs:
*       THREADINFO * info: 
* Result: void
*       
* Effect: 
*       Sets the thread info
****************************************************************************/

void CGraph::SetThreadInfo(THREADINFO * info)
    {
     ASSERT(info->GetIndex() < MAX_THREADS);
     if(info->GetIndex() >= MAX_THREADS)
        return;
     lines[info->GetIndex()].SetThreadInfo(info);
    } // CGraph::SetThreadInfo
